/*$6 ***************************************************************************
Copyright (C) <2007> <Sven Trenkel, Rene Weller, Gabriel Zachmann>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.
 ******************************************************************************* */

/**
 *  \ Pqp \
 */

/**
 *  @class Wrapper for Pqp-Library
 *  @author Sven Trenkel, Rene Weller, Gabriel Zachmann
 */

/*
 * Includes ;
 */
 
#include <OpenSG/OSGConfig.h>
#include <OpenSG/OSGBaseTypes.h>
#include <OpenSG/OSGNode.h>
#include <OpenSG/OSGVector.h>
#include <OpenSG/OSGMatrix.h>
#include <OpenSG/OSGGeometry.h>
#include <OpenSG/OSGGeoProperty.h>
#include <OpenSG/OSGGeoFunctions.h>
#include <OpenSG/OSGTriangleIterator.h>
#include <OpenSG/OSGBoxVolume.h>

#include <PQP.h>

#include "../Bench.h"

/*
 * Defines ;
 */
 
#define getGeom( x )    ( osg::GeometryPtr::dcast( (x)->getCore() ) )

using osg::     Vec3f;
using osg::     Pnt3f;

PQP_Model       **Obj;

/*
 * Forward Declarations ;
 */
 
static void     makeObject( int num, osg::NodePtr Node[] );
static int      doCollide( int num, osg::Matrix m[] );
static double   doDistance( int         num,
                            osg::Matrix m[],
                            osg::Pnt3f  *pos1,
                            osg::Pnt3f  *pos2 );
static void     shutdown();

static tAPI     API = { makeObject, doCollide, doDistance, shutdown };


/* -----------------------------------------------------------------------------
 *  \brief Init module and register it to the benchmarking application
 ------------------------------------------------------------------------------- */
#ifdef _WIN32
extern "C"
__declspec( dllexport )
std::vector<Algorithm *> *
initModule ( void )
#else
extern "C"
{
std::vector<Algorithm *>    *initModule( void );
}
std::vector<Algorithm *>    *
initModule ( void )
#endif
{
    std::vector<Algorithm *>    *ret = new std::vector < Algorithm * > ();
    ret->clear();
    Algorithm   *pqp = new Algorithm();
    pqp->shortName = "pqp";
    pqp->name = "PQP";
    pqp->api = API;
    ret->push_back( pqp );

    return ret;
}
PQP_CollideResult   cres;


/* -----------------------------------------------------------------------------
 *  \brief Collision check
 *  \param m transformation matrix of moving object
 *  \param i number of transformed object (has no effect in this version)
 ------------------------------------------------------------------------------- */

static int doCollide( int num, osg::Matrix m[] )
{
    /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
    PQP_REAL  rotMatrix1[3][3] = { { m[0][0][0], m[0][1][0], m[0][2][0] }, { m[0][0][1], m[0][1][1], m[0][2][1] }, { m[0][0][2], m[0][1][2], m[0][2][2] } };
    PQP_REAL  transVector1[3] = { m[0][3][0], m[0][3][1], m[0][3][2] };

    PQP_REAL  rotMatrix2[3][3] = { { 1, 0, 0 }, { 0, 1, 0 }, { 0, 0, 1 } };
    PQP_REAL  transVector2[3] = { 0, 0, 0 };
    /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

    PQP_Collide( &cres, rotMatrix1, transVector1, Obj[0], rotMatrix2,
                 transVector2, Obj[1], PQP_FIRST_CONTACT );
    return cres.Colliding();
}

PQP_DistanceResult  dres;

/* -----------------------------------------------------------------------------
 *  \brief Calculate distance between the moving- and the fixed object
 *  \param num # objects, has no effect in this version
 *  \param m transformation matrix of moving object
 *  \param pos1 pos2 closest points of the two objects
 ------------------------------------------------------------------------------- */

static double doDistance( int         num,
                          osg::Matrix m[],
                          osg::Pnt3f  *pos1,
                          osg::Pnt3f  *pos2 )
{
    /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
    PQP_REAL  rotMatrix1[3][3] = { { m[0][0][0], m[0][1][0], m[0][2][0] }, { m[0][0][1], m[0][1][1], m[0][2][1] }, { m[0][0][2], m[0][1][2], m[0][2][2] } };
    PQP_REAL  transVector1[3] = { m[0][3][0], m[0][3][1], m[0][3][2] };

    PQP_REAL  rotMatrix2[3][3] = { { 1, 0, 0 }, { 0, 1, 0 }, { 0, 0, 1 } };
    PQP_REAL  transVector2[3] = { 0, 0, 0 };
    /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

    PQP_Distance( &dres, rotMatrix1, transVector1, Obj[0], rotMatrix2,
                  transVector2, Obj[1], 0, 0 );
    if( pos1 )
    {
        const PQP_REAL  *Pnt = dres.P1();
        *pos1 = osg::Pnt3f( Pnt[0], Pnt[1], Pnt[2] );
    }

    if( pos2 )
    {
        const PQP_REAL  *Pnt = dres.P2();
        *pos2 = osg::Pnt3f( Pnt[0], Pnt[1], Pnt[2] );
    }

    return dres.Distance();
}

/* -----------------------------------------------------------------------------
 *  \brief Add an object to the collision detection library
 ------------------------------------------------------------------------------- */
static PQP_Model *addObject( osg::GeometryPtr geo )
{
    PQP_Model               *ret;
    osg::TriangleIterator   ti = geo->beginTriangles(),
    ti_end = geo->endTriangles();

    ret = new PQP_Model;
    ret->BeginModel();
    for( int t = 0; ti != ti_end; ++ti, ++t )
    {
        PQP_REAL  PosArray[3][3];
        Pnt3f   Pos[3];
        for( int i = 0; i < 3; ++i )
        {
            Pos[i] = ti.getPosition( i );
            PosArray[i][0] = Pos[i][0];
            PosArray[i][1] = Pos[i][1];
            PosArray[i][2] = Pos[i][2];
        }        
        ret->AddTri( PosArray[0], PosArray[1], PosArray[2], t );
    }

    ret->EndModel();
    return ret;
}

/* -----------------------------------------------------------------------------
 *  \brief Make the OSG-nodes collidable
 *  \param num # of objects
 *  \param Node OSG-Nodes of the objects
 ------------------------------------------------------------------------------- */
static void makeObject( int num, osg::NodePtr Node[] )
{
    Obj = new PQP_Model *[num + 1];
    Obj[num] = NULL;
    for( int i = 0; i < num; ++i )
    {
        Obj[i] = addObject( getGeom( Node[i] ) );
    }
}

static void shutdown()
{
    if ( !Obj ) return;
    for ( int i = 0; Obj[i]; ++i )
    {
        delete Obj[i];
    }
    delete[] Obj;
    Obj = NULL;
}
